# -*- coding: utf-8 -*-

# Copyright (C) 2010-2011 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2go is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
A recommended X2go session clean up helper function.

"""

import gevent
import paramiko
import threading

# Python X2go modules
import guardian
import rforward
from defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS

if _X2GOCLIENT_OS == 'Windows':
    import xserver
    import pulseaudio

def x2go_cleanup(e=None, threads=None):
    """\
    For every Python X2go application you write, please make sure to 
    capture the C{KeyboardInterrupt} and the C{SystemExit} exceptions and 
    call this function if either of the exceptions occurs.

    Example::

        import x2go

        try:
            my_x2goclient = x2go.X2goClient(...)

            [... your code ...]

            sys.exit(0)
        except (KeyboardInterrupt, SystemExit):
            x2go.x2go_cleanup()

    @param e: if L{x2go_cleanup} got called as you caught an exception in your code this can be the 
        C{Exception} that we will process at the end of the clean-up (or if clean-up failed or was not 
        appropriate)
    @type e: C{exception}
    @param threads: a list of threads to clean up
    @type threads: C{list}

    """
    try:
        if threads is None:
            threads = threading.enumerate()
        else:
            threads = threads

        # stop X2go reverse forwarding tunnels
        for t in threads:
            if type(t) == rforward.X2goRevFwTunnel:
                t.stop_thread()
                del t

        # stop X2go paramiko transports used by X2goTerminalSession objects
        for t in threads:
            if type(t) == paramiko.Transport:
                if hasattr(t, '_x2go_session_marker'):
                    t.stop_thread()
                    del t

        # on Windows: stop the XServer that we evoked
        if _X2GOCLIENT_OS == 'Windows':
            for t in threads:
                if type(t) == xserver.X2goXServer:
                    t.stop_thread()
                    gevent.sleep(1)
                    del t

        # on Windows: stop the PulseAudio daemon that we evoked
        if _X2GOCLIENT_OS == 'Windows':
            for t in threads:
                if type(t) == pulseaudio.X2goPulseAudio:
                    t.stop_thread()
                    gevent.sleep(1)
                    del t

        if e is not None:
            raise e

    except KeyboardInterrupt:
        # do not allow keyboard interrupts during Python X2go cleanup
        pass
    except SystemExit:
        # neither do we allow SIGTERM signals during cleanup...
        pass
